open util/ordering[Time] 
open util/ordering[A] 

sig Time {}

// "A" is ordered. We will perform a 
// binary search of A. 
sig A {}

// Pair each atom of A with an integer.
// That is, create an indexed A (i.e., an array).
one sig Array {
    items: Int lone -> lone A
} 
{
  // The values of A are associated with (indexed by) non-negative integers
  no i: Int - {i: Int | i >= 0} | some i.items
  // All values of A are indexed
  all a: A | some items.a
  // The indexes are consecutive, no gaps in the indexes
  all i: Int - {i: Int | i < 1}  | some i.items => some prev[i].items
  // If "a" is a value of A, then the index of a.prev is less than the index of "a".
  all a: A - first | items.(a.prev) < items.a
}

// This is how we model the variables used in
// the binary search algorithm: each variable 
// is time-varying. For example, at Time=0, the
// variable "low" has the array's smallest index (0).
// At Time=1, it might have the value mid+1, and
// so forth. So, each array index (low, high, mid) is
// paired to a Time. 
// midValue is the value of A[mid]. It also is time-varying.
// 
one sig Variables {
    low: Int lone -> Time,
    high: Int lone -> Time,
    mid: Int lone -> Time, 
    midValue: A lone -> Time,
    answer: Int lone -> Time,
    searchingFor: A  // This variable is just to help with debuggging
}

// This is a model of the binary search algorithm.
// Binary search the array for key.
pred binarySearch [key: A] { 
    Variables.searchingFor = key
    Variables.low.first = 0  // int low = 0;
    Variables.high.first = minus[#A, 1]  // int high = a.length - 1;
    no Variables.answer.first  // No answer at time 0
    all t: Time - first |
        some Variables.answer.(t.prev) => // If the answer was recorded at the previous time step, then ...
        {
            (Variables.answer.t  = Variables.answer.(t.prev))  // ... record it at each succeeding time step
            (Variables.high.t = Variables.high.(t.prev))  // No change to high
            (Variables.low.t = Variables.low.(t.prev)) // No change to low
            (Variables.mid.t = Variables.mid.(t.prev))  // No change to mid
            (Variables.midValue.t = Variables.midValue.(t.prev))  // No change to midValue
        }
        else   // Still searching for an answer ...
        {
            Variables.low.(t.prev) =< Variables.high.(t.prev) =>   // while (low <= high) 
            {
                Variables.mid.t = div[add[Variables.low.(t.prev), Variables.high.(t.prev)], 2]  // int mid = (low + high) / 2;
                Variables.midValue.t = Array.items[Variables.mid.t]  // int midVal = a[mid];
                Variables.midValue.t  in key.prevs =>  // if (midVal < key)
                {
                     Variables.low.t = plus[Variables.mid.t, 1] // low = mid + 1
                     Variables.high.t = Variables.high.(t.prev)  // No change to high
                     no Variables.answer.t  // No answer yet
                }
                else Variables.midValue.t  in key.nexts =>   // else if (midVal > key)
                {
                    Variables.high.t = minus[Variables.mid.t, 1]  // high = mid - 1;
                    Variables.low.t = Variables.low.(t.prev)  // No change to low
                    no Variables.answer.t  // No answer yet
                }
                else  // Variables.midValue.t = key  (i.e., found the answer!) 
                {
                    Variables.answer.t = Variables.mid.t 
                    Variables.high.t = Variables.high.(t.prev)  // No change to high
                    Variables.low.t = Variables.low.(t.prev)  // No change to low
                    no Variables.answer.{t': Time | t' in t.prevs}  // No answer recorded in previous time steps
                }
             }  // It's not the case that: while (low <= high)
             else  // key not found
             {
                    Variables.answer.t = add[1, minus[0, Variables.low.(t.prev)]]  // return -(low + 1);
                    Variables.high.t = Variables.high.(t.prev)  // No change to high
                    Variables.low.t = Variables.low.(t.prev)  // No change to low
                    no Variables.answer.{t': Time | t' in t.prevs}  // No answer recorded in previous time steps
             }
        }
}
run { some a: A | binarySearch [a] } for 10 but 6 Int
run { some x: univ | binarySearch [x] } for 10 but 6 Int

assert binarySearch_always_finds_the_answer {
    all a: A |
        binarySearch[a] =>
            some t: Time | Array.items[Variables.answer.t] = a
}
// The range of Int is -15 to 15, which is not large enough, e.g.
// suppose low is 7 and high is 9, then low + high = -16 due to overflow
check binarySearch_always_finds_the_answer for 10 but 5 Int
