/**************************************
          Components of XMI 
***************************************/

sig Element {
    name: Name,
    aggregation: lone Aggregation,
    children: set Element
}
one sig Root extends Element {}

sig Name {}
one sig SystemsContext extends Name {}

sig Aggregation {}

/**************************************
          Constraints on trees
***************************************/

// The Element signature defines a set of Element. 
// We want every value in that set to be part of 
// the tree.
fact No_disconnected_elements {
    all e: Element |
        (e = Root) or (e in Root.^children)
}

// An element cannot be a descendent of itself.
fact No_loops {
    no e: Element | e in e.^children
}

// Each element has one parent.
fact Each_element_has_one_parent {
    no disj e, e', e'': Element | 
        (e in e'.children) and (e in e''.children) 
}

/**************************************
          Constraints on XMI trees
***************************************/

// There must be exactly one element with name 'SystemsContext'
fact One_element_named_SystemsContext {
    one e: Element | (e.name = SystemsContext)
}

/**************************************
          Components of GraphML 
***************************************/

sig Node {
    name: Name,
    represents: Element
}

sig Edge {
    source: Node,
    target: Node 
}

one sig GraphML {
    nodes: set Node,
    edges: set Edge
}

// A helper signature. It contains all the elements for
// which we need to create a node.
one sig Relevant {
    elements: set Element
} {
    elements = {e: Element | (e in SystemsContext.~(Element <: name).*children) and 
                                                      ((e.name = SystemsContext) or (e.aggregation != none))}
}

/**************************************
          Constraints on GraphML
***************************************/

// The GraphML must contain a node for the 
// SystemsContext element and a 
// node  for each aggregation element 
// that descends  from the SystemsContext
// element 
fact GraphML_nodes {
    #GraphML.nodes = #Relevant.elements
    all e: Relevant.elements |
        one n: Node | 
            (n.name = e.name) and (n.represents = e) and (n in GraphML.nodes)
}

// The GraphML must contain an edge for each aggregation 
// element to the SystemsContext element. That is, the
// source node is the aggregation element, the target node
// is the SystemsContext element. 
fact GraphML_edges {
    #GraphML.edges = minus[#Relevant.elements,1]
    let targetNode = SystemsContext.~(Node <: name) |
        all sourceNode: Node - targetNode |
            some e: Edge | 
                (e.target = targetNode) and (e.source = sourceNode) and (e in GraphML.edges)
}

// The Node signature defines a set of Node. 
// We want every value in that set to be part of 
// GraphML.
fact No_disconnected_nodes {
    all n: Node |
        n in GraphML.nodes
}

// The Edge signature defines a set of Edge. 
// We want every value in that set to be part of 
// GraphML.
fact No_disconnected_edges {
    all e: Edge |
        e in GraphML.edges
}

pred Show {
    #Element = 4
    no disj e, e': Element | e.name = e'.name
   #(SystemsContext.~(Element <: name)).children > 1
}

run Show for 4

// This should yield a counterexample because aggregation
// elements that do not descend from SystemsContext do not
// have a corresponding Node.
assert There_is_a_node_for_every_aggregation_element {
    all e: aggregation.Aggregation |
        some n: Node |
            (n.represents = e) and (n.name = e.name)
}
check There_is_a_node_for_every_aggregation_element

// This should not yield a counterexample because we
// require a node corresponding to the SystemsContext element.
assert There_is_one_SystemsContext_node {
    one e: Element |
        (e.name = SystemsContext) and (e in Root.*children) and
       one n: Node |
            (n.name = SystemsContext) and (n.represents = e)
}
check There_is_one_SystemsContext_node 

// This should not yield a counterexample because nodes
// must correspond to either the SystemsContext element
// or to an aggregation element.
assert There_is_no_node_for_an_element_that_is_not_SystemsContext_or_an_aggregation_element {
    no n: Node |
        (n.represents.name != SystemsContext) and (n.represents.aggregation = none)
}
check There_is_no_node_for_an_element_that_is_not_SystemsContext_or_an_aggregation_element

// Given the GraphML, the only data visible should
// be the data in each Node and each Edge.
//  This should yield counterexamples because
// each Node references an Element in the XMI
// tree and once at an Element in the tree we 
// can traverse to all descendent elements and
// access their data.
// See http://alloytools.org/quickguide/meta.html
// There is a meta capability that allows you to 
// 'iterate' over the fields of an Atom.
// The value that each Node maps to must be a Name.
// Thus, the values in all fields of Node must be in Name.
assert No_data_leaks_from_Node {
    all f : Node$.subfields| f.value[Node] in Name
}
check No_data_leaks_from_Node

// The values that each Edge maps to must be a Node.
// Thus, the values in all fields of edge must be in Node.
assert No_data_leaks_from_Edge {
    all f : Edge$.subfields| f.value[Edge] in Node
}
check No_data_leaks_from_Edge
