module Common

// Given a set A and a relation R in A, R is reflexive 
// if and only if all the ordered pairs of the form (x, x) 
// are in R for every x in A.
pred reflexive (A: set univ, R: univ->univ) {
    all x: A | (x -> x) in R
}

// A relation which fails to be reflexive is called nonreflexive.
pred nonreflexive (A: set univ, R: univ->univ) {
    some x: A | (x -> x) not in R
}

// A relation which contains no ordered pair (x, x)
// with identical first and second members is 
// called irreflexive.
pred irreflexive (A: set univ, R: univ->univ) {
    no x: A | (x -> x) in R
}

// Given a set A and a relation R in A, R is symmetric
// if and only if for every ordered pair (x, y) in R, the
// pair (y, x) is also in R.
pred symmetric (A: set univ, R: univ->univ) {
    all disj x, y: A | ((x -> y) in R) => ((y -> x) in R)
}

// If for some (x, y) in R, the pair (y, x) is not in R
// then R is nonsymmetric.
pred nonsymmetric (A: set univ, R: univ->univ) {
    some x, y: A | ((x -> y) in R) and ((y -> x) not in R)
}

// Given a set A and a relation R in A, R is asymmetric
// if it is never the case that for any (x, y) in R, the
// pair (y, x) is in R.
pred asymmetric (A: set univ, R: univ->univ) {
    all x, y: A | ((x -> y) in R) => ((y -> x) not in R)
}

// Given a set A and a relation R in A, R is transitive
// if and only if for all ordered pairs (x, y) and (y, z) in R,
// the pair (x, z) is also in R.
pred transitive (A: set univ, R: univ->univ) {
    all x, y, z: A | (((x ->y) in R) and ((y -> z) in R)) => ((x -> z) in R)
}

// Given a set A and a relation R in A, if for no pairs 
// (x, y) and (y, z) in R, the pair (x, z) is in R, the 
// relation is intransitive.
pred intransitive (A: set univ, R: univ->univ) {
    all  x, y, z: A | (((x -> y) in R) and ((y -> z) in R)) => ((x -> z) not in R)
}

// If a relation fails to meet the definition of transitivity, 
// it is nontransitive. So, there may be transitivity in the 
// relation, but there is at least one (x,y) and (y,z) that 
// does not have (x,z).
pred nontransitive (A: set univ, R: univ->univ) {
    some  x, y, z: A | (((x -> y) in R) and ((y -> z) in R)) => ((x -> z) not in R)
}

// A relation R in A, is connected if and only if for every
// two distinct elements x and y in A, (x, y) in R or (y, x)
// in R or both.
pred connected (A: set univ, R: univ->univ) {
    all disj x, y: A | ((x -> y) in R) or ((y -> x) in R)
}

// If a relation fails to meet the definition of connected, 
// it is nonconnected. So, there are two distinct elements 
// x and y in A, such that (x, y) and (y, x) are not in R.
pred nonconnected (A: set univ, R: univ->univ) {
     some disj x, y: A | ((x -> y) not in R) and ((y -> x) not in R)
}

// Given a set A and a relation R in A, R is anti-symmetric
// if whenever both (x, y) and (x, y) are in R, then x = y.
pred antisymmetric (A: set univ, R: univ->univ) {
    all x, y: A | ((x -> y) in R) and ((y -> x) in R) => (x = y)
}

// A binary relation that maps at most one atom 
// to each atom is injective.
pred injective (A: set univ, R: univ->univ) {
    no disj  x, y, z: A | ((y -> x) in R) and ((z -> x) in R)
}

// A binary relation that maps each atom to at 
// most one other atom is functional and is 
// called a function. 
pred functional (A: set univ, R: univ->univ) {
    no disj  x, y, z: A | ((x -> y) in R) and ((x -> z) in R)
}
